/*---------------------------------------------------------------------------------------------*/
/* sourcefile: IRControl.c              date:       October 06, 2012                           */
/*                                      author(s):  M.Eckart/B.Frenzel                         */
/*                                                  martin.eck@me.com/bernhard.frenzel@me.com  */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Change index:                                                                               */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Program function:                                                                           */
/*    Software to activate an IR diode connected to output OUTPINS in dependance of an input   */
/*    pulse available at the input pin INPINS.                                                 */
/*    o input pulse (active high) with pulse length DOWN, MID or UP. the nominal pulse length  */
/*      is 1 ms for DOWN, 1.5 ms for MID and 2.0 ms for UP.                                    */
/*    o a Mealy state machine activates the IR diode to initiate video/picture funcionality    */
/*      on a Sony NEX-5N camera. The codes are programmed with the camera codes                */
/*      IRCMD_ADDRESS, IRCMD_SHUTTER, IRCMD_SHUTTER2S (unused), IRCMD_VIDON and IRCMD_VIDOFF.  */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Device(s):                          Atmel ATtiny25/45/85 or equivalent                      */
/*                                                                                             */
/*                                                                                             */
/* Programming Environment:                                                                    */
/*                          IDE        o Eclipse with AVR Eclipse Plugin, see eclipse.org      */
/*                          Software   o Arduino software environment, see www.arduino.cc      */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Timing:                             processor running @ 8 MHz                               */
/*                          By default, the devices ATtiny25/45/85 are shipped with the CKDIV8 */
/*                          fuse programmed, resulting in a system clock prescaler of 8.       */
/*                          So the default internal system clock will be 8 MHz/8 = 1 MHz,      */
/*                          see [1], p. 31,  6.2.7, NOT 8 MHz AS REQUIRED!                    */
/*                                                                                             */
/*                          !!! HENCE, FOR PROPER FUNCTION THE CLKDIV8 FUSEBIT SHALL BE !!!    */
/*                          !!! REPROGRAMMED TO 1 (=UNPROGRAMMED)                       !!!    */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Function/Signals:                                                                           */
/*                                                                                             */
/*   void IRsequence(uint8_t IRcmd)                                                            */
/*    Function sending the  Sony NEX-5N camera specific IR commands, coded in IRcmd. To        */
/*    control other cameras, the sequences in this function shall be adapted to the            */
/*    camera specific sequences.                                                               */
/*                                                                                             */
/*   ISR(INT0_vect)                                                                            */
/*    Interrupt Service Routine (ISR): will always be called when the external INT0 interrupt  */
/*    is activated. The ISR will determine the pulse width of the input INT0 line:             */
/*    On a rising edge of the INT0 line, the internal 8-bit counter will be reset to zero      */
/*    and will be restarted. On a falling edge, the counter is stopped and the counter value   */
/*    is stored in the global pulse width variable PW.                                         */
/*                                                                                             */
/*   int main(void)                                                                            */
/*    Main routine with                                                                        */
/*       a) processor initialization                                                           */
/*       b) test against new command received                                                  */
/*       c) dispatching according Mealy state diagram                                          */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Acknowledgements:                                                                           */
/*   The code in the function IRsequence is based on the author's work MasterFX                */
/*   (MasterFX@gmx.de), for more information see the thread in the dsrl-formum.de:             */
/*   http://www.dslr-forum.de/showthread.php?p=4321848                                         */
/*   Special thanks to MasterFX for the idea and the code.                                     */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Copyright/Warranty:                                                                         */
/*   This software underlies the GNU General Public License (GNU GPLv3), see:                  */
/*   http://www.gnu.org/licenses/gpl.html                                                      */
/*                                                                                             */
/*   As indicated, there is absolutely no warranty for the work and this software. The         */
/*   authors are not liable to you for damages, including any general, special, incidental     */
/*   or consequential damages arising out of the use or inability to use the program           */
/*   (including but not limited to loss of data or data being rendered inaccurate or losses    */
/*   sustained by you or third parties or afailure of the program to operate with any other    */
/*   programs), even if such holder or other party has been advised of the possibility of      */
/*   such damages (see  16, GNU GPLv3).                                                       */
/*                                                                                             */
/*                                                                                             */
/*                                                                                             */
/* Literature:                                                                                 */
/* [1] Atmel Corp.: 8-bit AVR Microcontroller with 2/4/8K Bytes In-System Programmable Flash,  */
/*     ATtiny 25/V,ATtiny 45/V, ATtiny 85/V, datasheet, Rev. 2586N-AVR-04/11, San Jose, 2011   */
/*                                                                                             */
/*---------------------------------------------------------------------------------------------*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdint.h>
#include <util/delay.h>

/*...........................................................................define section */
#define PHOTO_INTERVAL 1000          /* time interval in msec between two photos            */

#define INPINS  (1<<PB2)              /* input pin(s):  PB2 = INT0                          */
#define OUTPINS ((1<<PB3) | (1<<PB4)) /* output pin(s): PB3, PB4                            */

                           /* counter adjustments                                           */
#define START_CNT  TCCR0B |=  (1<<CS02); /* start counter with prescaled clock to f_IO/256  */
#define STOP_CNT   TCCR0B &= ~(1<<CS02); /* start counter with prescaled clock to f_IO/256  */

                           /* nominal 8 bit timer count for down/mid/up command             */
#define DOWN 32            /* stick down: nominal timer count when 1.0 ms pulse is detected */
#define MID  47            /* stick mid:  nominal timer count when 1.5 ms pulse is detected */
#define UP   63            /* stick mid:  nominal timer count when 2.0 ms pulse is detected */
#define VIDONDELAY 25      /* delay in ticks before state transition VIDDELAY->VIDON        */

                           /* lower/upper limits to detect DOWN/MID/UP command:             */
#define DCNT 3             /* counter tolerance DOWN/MID/UP +/- DCNT to detect valid pulse  */
#define LLIM_DOWN  DOWN-DCNT  /* low command if LLIM_DOWN <= counter value <= ULIM_DOWN     */
#define ULIM_DOWN  DOWN+DCNT
#define LLIM_MID    MID-DCNT  /* mid command if LLIM_MID  <= counter value <= ULIM_MID      */
#define ULIM_MID    MID+DCNT
#define LLIM_UP      UP-DCNT  /* up command  if  LLIM_UP  <= counter value <= ULIM_UP       */
#define ULIM_UP      UP+DCNT

                            /* IR commands for Sony NEX-5N:                                 */
#define IRCMD_ADDRESS    0x1E3A       /* IR command address                                 */
#define IRCMD_SHUTTER    0x2D         /* IR command for immediate shutter activation        */
#define IRCMD_SHUTTER2S  0x37         /* IR command for shutter activation with 2 sec delay */
#define IRCMD_VIDON      0x48         /* IR command for video (de)activation                */
#define IRCMD_VIDOFF     0x48         /* IR command for video (de)activation                */

                            /* Interrupt handling macros (ATtiny specific)                  */
#define MASK_INT0        (GIMSK &= ~(1<<INT0)) /* mask   INTO IRQ                           */
#define UNMASK_INT0      (GIMSK |=  (1<<INT0)) /* unmask INT0 IRQ                           */
#define CHG_EDGEDETECT   (MCUCR ^= (1<<ISC00)) /* change edge detect rising<->falling edge  */
#define FALLING_EDGE     ((MCUCR & (1<<ISC00)) == 0) /* true if falling edge detected       */
#define RISING_EDGE      ((MCUCR & (1<<ISC00)) == 1) /* true if rising  edge detected       */
#define ADJ_RISING_EDGE  (MCUCR |= (1<<ISC01) | (1<<ISC00)) /* adjust rising edge INTO      */

/* Macros for IR LED */
#define LED_ON()        PORTB |=  OUTPINS                         /* write 1 to output pins */
#define LED_OFF()       PORTB &= ~OUTPINS                         /* write 0 to output pins */
    /* pulse definitions for start sequence frame:                                          */
    /* set output high (=1), wait 12.5 seconds, set output low (=0), wait 12.5 seconds    */
#define PULSE_40k()     LED_ON(); _delay_us(12.5); LED_OFF(); _delay_us(12.5)


/*.....................................................................own type definitions */
typedef enum        /* states of main Mealy state machine                                   */
{
   IDLE    = 0,     /* IDLE state:    do nothing                                            */
   SHUTTER,         /* SHUTTER state: take a picture                                        */
   VIDDELAY,        /* VIDDELAY state: wait VIDONDELAY ticks before VIDON state             */
   VIDON,           /* VIDON state:   start video recording                                 */
   VIDWAIT,         /* VIDWAIT state: wait while video recording is on                      */
   VIDOFF           /* VIDOFF state:  stop video recording                                  */
} state_t;

typedef enum        /* command from ppm channel (remote stick)                              */
{
   UNKNOWN  = 0,    /* unknown command                                                      */
   CMDDOWN,         /* stick down command                                                   */
   CMDMID,          /* stick mid command                                                    */
   CMDUP            /* stick up command                                                     */
} cmd_t;

/*..................................................................global variable section */
volatile uint8_t READY_FLAG = 0;         /* READY_FLAG = 0 for rising edge, = 1 for falling */
volatile uint8_t PW = 0;                 /* pulse width in multiples of counter period time */

/*...........................................................functional declaration section */
void IRsequence(uint8_t);                /* function to send IR commands                    */



/*.............................................................................main routine */
int main(void){
	/* variable declarations */
   state_t  IRState  = IDLE;             /* states of Mealy state machine                   */
   cmd_t    cmd      = UNKNOWN;          /* command from ppm channel (remote stick)         */
   uint8_t viddelay = 0;                /* delay timer for video on start                  */

/*...........................................................................initialization */
   DDRB   |= OUTPINS;                              /* set output port pins for IR light     */
   PORTB  |= INPINS;                               /* enable pullup for input port pins     */

   ADJ_RISING_EDGE;        /* Initially start with a rising edge of INT0 to generate IRQ    */
   UNMASK_INT0;            /* Unmask External Interrupt INT0                                */
   sei();	           /* Enable all interrupts now: set I-bit in SREG to 1             */

/*..................................................................... main (endless) loop */


   while(1){
   /* __________________________ test against new valid command ___________________________ */
   if (READY_FLAG) {                                   /* check for valid command:          */
      if ((PW >= LLIM_MID) & (PW <= ULIM_MID))
         cmd = CMDMID;                                 /* stick in middle position          */
      else if ((PW >= LLIM_DOWN) & (PW <= ULIM_DOWN))
         cmd = CMDDOWN;                                /* stick in down position            */
      else if ((PW >= LLIM_UP)   & (PW <= ULIM_UP))
         cmd = CMDUP;                                  /* stick in up position              */
      else
         cmd = UNKNOWN;                                /* position not defined              */

      READY_FLAG = 0;                                  /* reset ready flag                  */
  }

  /* ______________________ dispatching according state diagram ___________________________ */
  if (cmd != UNKNOWN) {    /* if valid command detected                                     */
      cli();	           /* Disable all interrupts now: clear I-bit in SREG to 0          */
      switch (IRState) {
           case IDLE:
     		  viddelay = 0;                 /* reset delay timer                            */
              if (cmd == CMDDOWN) {
                 IRState    = VIDDELAY;     /* new state = VIDDELAY                         */
              } else if (cmd == CMDUP) {
                 IRState    = SHUTTER;      /* new state = SHUTTER                          */
              } else {
                 IRState    = IDLE;         /* new state = IDLE                             */
              }
              break;

           case SHUTTER:
              if (cmd == CMDMID) {
                 IRState = IDLE;           /* new state = IDLE                              */
              } else {
                IRsequence(IRCMD_SHUTTER); /* make photo                                    */
                _delay_ms(PHOTO_INTERVAL); /* wait defined interval in ms for next photo    */
                IRState = SHUTTER;         /* new state = SHUTTER                           */
              }
              break;

           case VIDDELAY:
              if (cmd == CMDDOWN) {
            	  viddelay++;              /* increment delay timer                         */
            	  if (viddelay >= VIDONDELAY) {
            		  IRState  = VIDON;    /* new state = VIDON                             */
            	  }
              } else {
            	  IRState = IDLE;          /* new state = IDLE                              */
              }
        	 break;

           case VIDON:
              IRsequence(IRCMD_VIDON);     /* turn on video                                 */
              IRState = VIDWAIT;           /* new state = VIDWAIT                           */
              break;

           case VIDWAIT:
              if (cmd == CMDMID) {
            	  IRState = VIDOFF;         /* new state = VIDOFF                            */
              } else {
                 IRState = VIDWAIT;        /* new state = VIDWAIT                           */
              }
              break;

            case VIDOFF:
              IRsequence(IRCMD_VIDOFF);   /* turn off video                                 */
              IRState = IDLE;             /* new state = IDLE                               */
              break;

           default:
              IRState = IDLE;         /* default state = IDLE                               */
              break;
       }                   /* end of switch                                                 */
       cmd = UNKNOWN;      /* command is unknown command for next cycle                     */
       sei();              /* Enable all interrupts now: set I-bit in SREG to 1             */
       }                   /* end of  if (cmd != UNKNOWN)                                   */
   }                       /* end of while(1) loop                                          */

	return 0;              /* should never reach this...                                    */
}



/*.............................................................function to send IR commands */
void IRsequence(uint8_t IRcmd) {
                                                           /* Variable declarations:        */
uint8_t i,j=0,k;                                           /* counter variables             */
uint16_t mask = 1;                                         /* mask for only one bit to send */
uint16_t address = IRCMD_ADDRESS;                          /* camera command                */

  	for(k = 0; k<5; k++){                              /* repeat 5 times                */
		mask = 1;
		                          /* Send Start-Burst = 96 times the start sequence */
		for(i = 0; i < 96; i++){
			PULSE_40k();
		}
		_delay_us(640);

		j=0;                      /* Send command with LSB first                    */
		while(j++ <7){
		   if(IRcmd & mask){
                      for(i = 0; i < 48; i++){
                         PULSE_40k();
                      }
                   } else {
                      for(i = 0; i < 24; i++){
                         PULSE_40k();
                      }
                   }
                   _delay_us(640);
                   mask <<= 1;
		}

		j=0;                      /* Send address with LSB first                    */
		mask = 1;
		while(j++<13){
                   if(address & mask){
                      for(i = 0; i < 48; i++){
                         PULSE_40k();
                      }
                   } else{
                      for(i = 0; i < 24; i++){
                         PULSE_40k();
                      }
                   }
                   _delay_us(640);
                   mask <<= 1;
		}
		_delay_ms(11);          /* Wait 11 msec                                    */
	}
} /* end of IRsequence */




/*..........................................................interrupt service routine (ISR) */
ISR(INT0_vect){
     if (FALLING_EDGE){   /* if (the IRQ was generated by a falling edge)                   */
        STOP_CNT;         /*    stop counter(s)                                             */
        PW = TCNT0;       /*    read pulse width from 8-bit counter 0                       */
        READY_FLAG = 1;   /*    set ready flag                                              */
     } else {             /* else raising edge generated the IRQ                            */
        STOP_CNT;         /*    stop counter(s)                                             */
        TCNT0      = 0;   /*    clear counter value                                         */
        START_CNT;        /*    start counter(s)                                            */
        READY_FLAG = 0;   /*    clear ready flag and wait for falling edge                  */
     }
     CHG_EDGEDETECT;      /* change from rising edge detect to falling and vice versa       */
}

/*...........................................................................end of program */
